#include <jni.h>

#ifdef _WIN32
#define PATH_SEPARATOR ';'
#else /* UNIX */
#define PATH_SEPARATOR ':'
#endif

static JNIEnv *env;
static JavaVM *jvm;

static jclass Amanda, Particle;
static jobject amanda, particle, aobj, eobj, sobj;
static const char *name;

void initJre(){

  jint res;
  JavaVMInitArgs vm_args;
  JavaVMOption options[4];

  options[0].optionString = "-Djava.compiler=JIT";          /* NONE disables JIT */
  options[1].optionString = "-Djava.class.path=mmc.jar";    /* user classes */
  options[2].optionString = "-Djava.library.path=.";        /* set native library path */
  options[3].optionString = "-verbose:none";                /* jni prints JNI-related messages */

  vm_args.version = JNI_VERSION_1_4;
  vm_args.options = options;
  vm_args.nOptions = 4;
  vm_args.ignoreUnrecognized = 1;

  res = JNI_CreateJavaVM(&jvm, (void **)&env, &vm_args);
  if (res < 0) {
    fprintf(stderr, "cannot create Java VM\n");
    exit(1);
  }

  fprintf(stderr, "jre started\n");

}


void stopJre(){

  (*jvm)->DestroyJavaVM(jvm);
  fprintf(stderr, "jre stopped\n");

}


void initMMC(char *options){

  jmethodID mid;
  jstring jstr;

  Amanda = (*env)->FindClass(env, "tfa/Amanda");
  if (Amanda == 0) {
    fprintf(stderr, "cannot find the main class\n");
    exit(1);
  }

  mid = (*env)->GetMethodID(env, Amanda, "<init>", "()V");
  if (mid == 0) {
    fprintf(stderr, "cannot find the constructor\n");
    exit(1);
  }

  amanda = (*env)->NewObject(env, Amanda, mid);
  if (amanda == 0) {
    fprintf(stderr, "cannot start the constructor\n");
    exit(1);
  }

  mid = (*env)->GetMethodID(env, Amanda, "setup", "(Ljava/lang/String;)V");
  if (mid == 0) {
    fprintf(stderr, "cannot find setup method\n");
    exit(1);
  }

  jstr = (*env)->NewStringUTF(env, options);
  if (jstr == 0) {
    fprintf(stderr, "Out of memory\n");
    exit(1);
  }

  (*env)->CallVoidMethod(env, amanda, mid, jstr);
  fprintf(stderr, "MMC initialized\n");

}


void propagate(char *name, double x, double y, double z, double theta, double phi, double e, double t){

  jmethodID pmid, amid;
  jstring jstr;

  // create new particle

  Particle = (*env)->FindClass(env, "mmc/Particle");
  if (Particle == 0) {
    fprintf(stderr, "cannot find the Particle class\n");
    exit(1);
  }

  pmid = (*env)->GetMethodID(env, Particle, "<init>", "(Ljava/lang/String;DDDDDDD)V");
  if (pmid == 0) {
    fprintf(stderr, "cannot find the Particle constructor\n");
    exit(1);
  }

  jstr = (*env)->NewStringUTF(env, name);
  if (jstr == 0) {
    fprintf(stderr, "Out of memory\n");
    exit(1);
  }

  particle = (*env)->NewObject(env, Particle, pmid, jstr, x*1.e2, y*1.e2, z*1.e2, 180-theta, phi<180?phi+180:phi-180, e*1.e3, t);
  if (particle == 0) {
    fprintf(stderr, "cannot start the Particle constructor\n");
    exit(1);
  }

  // propagate the particle

  amid = (*env)->GetMethodID(env, Amanda, "propagate", "(Lmmc/Particle;)[Lmmc/Particle;");
  if (amid == 0) {
    fprintf(stderr, "cannot find the propagate method\n");
    exit(1);
  }

  aobj=(*env)->CallObjectMethod(env, amanda, amid, particle);
  if (aobj == 0) {
    fprintf(stderr, "cannot run the propagate method\n");
    exit(1);
  }
}


int getNum(){
  return (*env)->GetArrayLength(env, aobj);
}

void startRead(int i){
  eobj=(*env)->GetObjectArrayElement(env, aobj, i);
}

void endRead(){
  (*env)->ReleaseStringUTFChars(env, sobj, name);
}

const char *getName(){
  sobj=(*env)->GetObjectField(env, eobj, (*env)->GetFieldID(env, Particle, "name", "Ljava/lang/String;"));
  return name=(*env)->GetStringUTFChars(env, sobj, NULL);
}

double getDouble(char *s, int i){
  switch(i){
  case 0: return (*env)->GetDoubleField(env, eobj, (*env)->GetFieldID(env, Particle, s, "D"));
  case 1: return (*env)->GetDoubleField(env, particle, (*env)->GetFieldID(env, Particle, s, "D"));
  case 2: return (*env)->GetDoubleField(env, amanda, (*env)->GetFieldID(env, Amanda, s, "D"));
  default: return 0;
  }
}

int getType(){
  return (*env)->GetIntField(env, eobj, (*env)->GetFieldID(env, Particle, "type", "I"));
}
