package mmc;

/**
 * class contains functions necessary for the calculation of decay
 */

public class Decay extends CrossSections implements DFunctionOfx{

    private FindRoot f;

    /**
     * creates internal references to p and m
     */

    public Decay(CrossSections cros){
	super(cros);
	f = new FindRoot(imaxs, iprec);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * this cross section describes decay
     */

    public double decay(){
	if(cros.cd<=0) return 0;
	return cros.cd/Math.max((p.p/p.m)*p.l*C, xres);
    }

    //----------------------------------------------------------------------------------------------------//

    public String out;

    /**
     * energy of the electron that results from the muon decay
     */

    public double e(double ernd, double arnd, double srnd){
	double emax, x0, f0, el, lm;
	if(p.type==2){
	    final double brmu=0.1737;
	    final double brel=0.1783+brmu;
	    final double brpi=0.1109+brel;
	    final double br2p=0.2540+brpi;
	    final double br3p=0.1826+br2p;
	    if(srnd<brmu){
		lm=Mmu;
		if(p.name.endsWith("+")) out="mu+";
		else if(p.name.endsWith("-")) out="mu-";
		else out="mu";
	    }
	    else if(srnd<brel){
		lm=Me;
		if(p.name.endsWith("+")) out=Output.AMASIM?"epair":"e+";
		else if(p.name.endsWith("-")) out=Output.AMASIM?"delta":"e-";
		else out=Output.AMASIM?"delta":"e";
	    }
	    else{
		if(srnd<brpi) lm=Mpi;
		else if(srnd<br2p) lm=Mrh;
		else if(srnd<br3p) lm=Ma1;
		else lm=Mrs;
		el=(Mtau*Mtau+lm*lm)/(2*Mtau);
		out=Output.AMASIM?"munu":"hadr";
		return el*(p.e/p.m)+Math.sqrt(el*el-lm*lm)*(p.p/p.m)*(2*arnd-1);
	    }
	}
	else{
	    lm=Me;
	    if(p.name.endsWith("+")) out=Output.AMASIM?"epair":"e+";
	    else if(p.name.endsWith("-")) out=Output.AMASIM?"delta":"e-";
	    else out=Output.AMASIM?"delta":"e";
	}
	emax=(p.m*p.m+lm*lm)/(2*p.m);
	x0=lm/emax;
	f0=x0*x0;
	f0=f0*x0-f0*f0/2;
	el=Math.max(f.findRoot(x0, 1, 0.5, this, f0+(0.5-f0)*ernd)*emax, lm);
	return el*(p.e/p.m)+Math.sqrt(el*el-lm*lm)*(p.p/p.m)*(2*arnd-1);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * function for electron energy calculation - interface to FindRoot
     */

    public double function(double x){
	double x2;
	x2=x*x;
	return x*x2-x2*x2/2;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * function for electron energy calculation - interface to FindRoot
     */

    public double dFunction(double x){
	return (3-2*x)*x*x;
    }

}
