package tfa;
import mmc.*;
import java.io.File;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.Vector;
import java.util.StringTokenizer;

/**
 * Implements muon/tau propagation through multiple media to/through the cylindrical detector.
 */

public class Amanda{

    private final static boolean lfix=true;
    private final static double LENGTH=800;
    private final static double RADIUS=400;
    private final static double BIG=6.40e7;

    private boolean SURF=false;
    private boolean FACE=false;
    private boolean USER=false;
    private boolean USFI=false;
    private boolean SDEC=false;
    private double zset=0;
    private String type;

    private int mediamax=100;
    private String mediadef=null;

    private double vcut=0.05;
    private double ecut=5.e2;
    private double vaux=-1.0;
    private double eaux=-1.0;
    private String med="Ice";
    private String muta="mu";
    private double elow=PhysicsModel.elow;
    private double ebig=PhysicsModel.ebig;

    private boolean timef=false;
    private boolean conti=false;
    private boolean lpmef=false;
    private boolean scatt=false;
    private boolean rfix=false;
    private boolean amasim=false;

    private int bspar=1;
    private int pncrs=1;
    private int pncbb=1;
    private int pncsh=1;

    private double crsci=1;
    private double crscb=1;
    private double crscp=1;
    private double crsce=1;
    private double crscd=1;
    private double rho=1;

    private int romb=5;
    private long seed=0;
    private boolean SEED=false;
    private boolean raw=false;
    private String tdir="";
    private String dtct="";

    private Propagate p1[], p2[], p3[];
    private double length=LENGTH, radius=RADIUS, surf=0;
    private double emax=0;
    private StringBuffer hist;
    private Vector userbf;
    private Vector I3hist;

    private int igen, gens=0, imax=0;
    private long events=0, tracks=0, vertices=0, missed=0;
    private String prnt;

    private String param="Amanda", options;
    private int fnu;
    private int fnm[];
    private double fnx[];
    private int medianum=1;
    private int spht[];
    private double sphz[];
    private double sphr[];

    //----------------------------------------------------------------------------------------------------//

    /**
     * Front-end for AMANDA, reads from and outputs to standard input/output in F2000 format.
     */

    public static void main(String[] args){
	Amanda A = new Amanda();
	A.mmcf2k(args);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * Initializes propagation for external applications, e.g.&nbsp;mmc-icetray (through jni).
     */

    public void setup(String args){
	int i=0, j;
	Vector parm = new Vector(Output.HISTSIZE);
	while((j=args.indexOf(" ", i))!=-1){
	    if(j>0) parm.addElement(args.substring(i, j));
	    i=j+1;
	}
	if(i<args.length()) parm.addElement(args.substring(i, args.length()));
	String parms[] = new String[parm.size()];
	for(i=0; i<parms.length; i++) parms[i]=(String)parm.elementAt(i);
	Output.I3flag=true; dtct=".icecube";
	setp(parms);
	I3hist = new Vector(Output.HISTSIZE);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * Propagates particles for the external applications. Returns an array of secondaries.
     * If "-user" option is used, fills user-block variables.
     */

    public Particle[] propagate(Particle p){
	if(!Output.I3flag) return null;
	Particle[] I3p;

	type=p.name;
	if(muta.equals(type) || (muta+"-").equals(type) || (muta+"+").equals(type)){
	    p.r=prop(p.x*1.e-2, p.y*1.e-2, p.z*1.e-2, 180-p.theta, p.phi<180?p.phi+180:p.phi-180, p.l*1.e-2, p.e*1.e-3, p.t*1.e9);
	    I3p = new Particle[I3hist.size()];
	    for(int i=0; i<I3hist.size(); i++) I3p[i]=(Particle)I3hist.elementAt(i);
	}
	else{
	    I3p = new Particle[0];
	}
	return I3p;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * This is the command-line option parser. Call with "-help" to list all options.
     */

    private boolean setp(String[] args){
	int i, j;

	int bnum=0;
	String pbad="";
	boolean pflag;

	double vcut[] = new double[mediamax];
	double ecut[] = new double[mediamax];
	double rho[] = new double[mediamax];
	String med[] = new String[mediamax];
	boolean conti[] = new boolean[mediamax];

	vcut[0]=this.vcut;
	ecut[0]=this.ecut;
	rho[0]=this.rho;
	med[0]=this.med;
	conti[0]=this.conti;

	String mediaopts="";
	int mediaseg=1;
	spht = new int[mediamax];
	sphz = new double[mediamax];
	sphr = new double[mediamax];
	boolean sphn[] = new boolean[mediamax];

	medianum=1;
	spht[0]=0;
	sphn[0]=true;

	for(int n=0; n<args.length; n++){
	    pflag=true;
	    if(args[n].equals("-help") || args[n].equals("-h") || args[n].equals("--help")){
		Output.out.println("\n"+
"This program propagates muons in "+med[0]+" to/through the detector\n"+
"Available options are: -length=[LENGTH of the detector volume in meters]\n"+
"                       -radius=[RADIUS of the detector volume in meters]\n"+
"                       -vcut=[value of vcut used for the 1st region]\n"+
"                       -ecut=[ecut in  MeV  used for the 2nd region]\n"+
"                       -medi=[medium name]\n"+
"                       -mediadef=[file with media definitions]\n"+
"                       -tau  propagate taus instead of muons\n"+
"                       -sdec enable stopped muon decay treatment\n"+
"                       -user     enable the mmc_en user line\n"+
"                       -user=[z] same, but record energy at z, not CPD\n"+
"                       -rdmc     enforce compliance with rdmc\n"+
"                       -amasim   turn on workarounds for amasim\n"+
"                       -time precise time of flight calculation\n"+
"                       -cont enable continuous loss randomization\n"+
"                       -scat enable Moliere scattering\n"+
"                       -lpm  enable lpm treatment\n"+
"                       -bs=[1-2]  bremsstrahlung: (1) kkp or (2) abb\n"+
"                       -ph=[1-4]  photonuclear: bb, bb+bs, allm, bm\n"+
"                       -bb=[bb/bs:1-4 3|4:bb|zeus, allm:1-2(91/7), bm:1]\n"+
"                       -sh=[1-2]  nuclear structure function: dutt/butk\n"+
"                       -c[i/b/p/e/d]=[cross section modifier, 0:disable]\n"+
"                       -rho=[multiplicative factor] for medium density\n"+
"                       -elow=[muon energy in GeV below which it is lost]\n"+
"                       -ebig=[upper bound in GeV of the paramet. tables]\n"+
"                       -surf=[h in meters]  propagate to the plane z=[h]\n"+
"                       -face  only if detector is on opposite side of it\n"+
"                       -romb=[number of interpolation points]\n"+
"                       -seed=[integer] sets random number generator seed\n"+
"                       -raw  save tables in raw format\n"+
"                       -tdir=[dir] specify directory for paramet. tables\n");
		return false;
	    }
	    else if(args[n].equals("-tau")){
		muta="tau";
	    }
	    else if(args[n].equals("-sdec")){
		SDEC=true;
	    }
	    else if(args[n].equals("-user")){
		USER=true;
	    }
	    else if(args[n].equals("-rdmc")){
		rfix=true;
	    }
	    else if(args[n].equals("-amasim")){
		amasim=true;
	    }
	    else if(args[n].equals("-time")){
		timef=true;
	    }
	    else if(args[n].equals("-cont")){
		conti[0]=true;
	    }
	    else if(args[n].equals("-scat")){
		scatt=true;
	    }
	    else if(args[n].equals("-lpm")){
		lpmef=true;
	    }
	    else if(args[n].equals("-face")){
		FACE=true;
	    }
	    else if(args[n].equals("-raw")){
		raw=true;
	    }
	    else if(args[n].startsWith("-length=")){
		try{
		    length=Double.parseDouble(args[n].substring(8));
		}catch(NumberFormatException error){
		    length=0;
		}
	    }
	    else if(args[n].startsWith("-radius=")){
		try{
		    radius=Double.parseDouble(args[n].substring(8));
		}catch(NumberFormatException error){
		    radius=0;
		}
	    }
	    else if(args[n].startsWith("-surf=")){
		try{
		    surf=Double.parseDouble(args[n].substring(6));
		}catch(NumberFormatException error){
		    surf=0;
		}
		SURF=true;
	    }
	    else if(args[n].startsWith("-vcut=")){
		try{
		    vcut[0]=Double.parseDouble(args[n].substring(6));
		}catch(NumberFormatException error){
		    vcut[0]=this.vcut;
		}
	    }
	    else if(args[n].startsWith("-ecut=")){
		try{
		    ecut[0]=Double.parseDouble(args[n].substring(6));
		}catch(NumberFormatException error){
		    ecut[0]=this.ecut;
		}
	    }
	    else if(args[n].startsWith("-user=")){
		try{
		    USER=true; USFI=true;
		    zset=Double.parseDouble(args[n].substring(6));
		}catch(NumberFormatException error){
		    zset=0;
		}
	    }
	    else if(args[n].startsWith("-bs=")){
		try{
		    bspar=(int)Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    bspar=1;
		}
	    }
	    else if(args[n].startsWith("-ph=")){
		try{
		    pncrs=(int)Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    pncrs=1;
		}
	    }
	    else if(args[n].startsWith("-bb=")){
		try{
		    pncbb=(int)Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    pncbb=1;
		}
	    }
	    else if(args[n].startsWith("-sh=")){
		try{
		    pncsh=(int)Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    pncsh=1;
		}
	    }
	    else if(args[n].startsWith("-ci=")){
		try{
		    crsci=Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    crsci=1;
		}
	    }
	    else if(args[n].startsWith("-cb=")){
		try{
		    crscb=Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    crscb=1;
		}
	    }
	    else if(args[n].startsWith("-cp=")){
		try{
		    crscp=Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    crscp=1;
		}
	    }
	    else if(args[n].startsWith("-ce=")){
		try{
		    crsce=Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    crsce=1;
		}
	    }
	    else if(args[n].startsWith("-cd=")){
		try{
		    crscd=Double.parseDouble(args[n].substring(4));
		}catch(NumberFormatException error){
		    crscd=1;
		}
	    }
	    else if(args[n].startsWith("-rho=")){
		try{
		    rho[0]=Double.parseDouble(args[n].substring(5));
		}catch(NumberFormatException error){
		    rho[0]=this.rho;
		}
	    }
	    else if(args[n].startsWith("-elow=")){
		try{
		    elow=Double.parseDouble(args[n].substring(6))*1.e3;
		}catch(NumberFormatException error){
		    elow=PhysicsModel.elow;
		}
	    }
	    else if(args[n].startsWith("-ebig=")){
		try{
		    ebig=Double.parseDouble(args[n].substring(6))*1.e3;
		}catch(NumberFormatException error){
		    ebig=PhysicsModel.ebig;
		}
	    }
	    else if(args[n].startsWith("-romb=")){
		try{
		    romb=(int)Double.parseDouble(args[n].substring(6));
		}catch(NumberFormatException error){
		    romb=5;
		}
	    }
	    else if(args[n].startsWith("-seed=")){
		try{
		    seed=Long.parseLong(args[n].substring(6));
		}catch(NumberFormatException error){
		    seed=0;
		}
		SEED=true;
	    }
	    else if(args[n].startsWith("-medi=")){
		med[0]=args[n].substring(6).replace('-', ' ');
	    }
	    else if(args[n].startsWith("-mediadef=")){
		mediadef=args[n].substring(10);
	    }
	    else if(args[n].startsWith("-tdir=")){
		tdir=args[n].substring(6)+"/";
	    }
	    else{
		bnum++;
		pbad+=(bnum>1?",":"")+" \""+args[n]+"\"";
		pflag=false;
	    }
	    if(pflag) param+=" "+args[n];
	}

	Output.err.println(Output.version);
	Output.err.println("Running \""+param+"\"");
	if(bnum>0){
	    param+=" (not used:"+pbad+")";
	    pbad=bnum==1?pbad+" is":"s"+pbad+" are";
	    Output.err.println("Warning: Parameter"+pbad+" not recognized");
	}

	if(length<=0){
	    Output.err.println("Warning: length is not a valid number");
	    length=LENGTH;
	}
	if(radius<=0){
	    Output.err.println("Warning: radius is not a valid number");
	    radius=RADIUS;
	}
	if(!SURF || surf==0) FACE=false;
	if(bspar<1 || bspar>2){
	    Output.err.println("Warning: bs is not a valid number");
	    bspar=1;
	}
	if(pncrs<1 || pncrs>4){
	    Output.err.println("Warning: ph is not a valid number");
	    pncrs=1;
	}
	if(((pncrs==1 || pncrs==2) && (pncbb<1 || pncbb>4)) ||
	    (pncrs==3 && (pncbb<1 || pncbb>2)) || (pncrs==4 && pncbb!=1)){
	    Output.err.println("Warning: bb is not a valid number");
	    pncbb=1;
	}
	if(((pncrs==1 || pncrs==2) && (pncsh!=1)) || ((pncrs>2) && (pncsh<1 || pncsh>2))){
	    Output.err.println("Warning: sh is not a valid number");
	    pncsh=1;
	}
	if(romb<2 || romb>6){
	    Output.err.println("Warning: romb is not a valid number");
	    romb=5;
	}

	if(mediadef!=null){
	    try{
		if((new File(mediadef)).exists()){
		    FileReader ifile = new FileReader(mediadef);
		    LineNumberReader inf = new LineNumberReader(ifile);
		    int mediacur;
		    String buf;
		    mediaopts="HI  Attaching the media definition file:\n";
		    while((buf=inf.readLine())!=null){
			mediaopts+="HI  ! "+buf+"\n";
			StringTokenizer st = new StringTokenizer(buf);
			if(!st.hasMoreTokens()) continue;
			String taux=st.nextToken();
			if(taux.charAt(0)=='#') continue;
			if(medianum==mediamax){
			    Output.err.println("Warning: Number of defined media will not exceed "+mediamax);
			    break;
			}
			if("all".equals(taux.toLowerCase())){
			    mediacur=0;
			    spht[mediacur]=0;
			}
			else if("sphere".equals(taux.toLowerCase())){
			    mediacur=medianum;
			    spht[mediacur]=1;
			    sphz[mediacur]=Double.parseDouble(st.nextToken());
			    sphr[mediacur]=Double.parseDouble(st.nextToken());
			    for(i=0; i<medianum; i++) if(spht[i]==1)
				if(sphz[i]==sphz[mediacur] && sphr[i]==sphr[mediacur]) break;
			    if(i==medianum){ medianum++; mediaseg+=2; }
			}
			else continue;
			sphn[mediacur]=Integer.parseInt(st.nextToken())>0?true:false;
			vcut[mediacur]=Double.parseDouble(st.nextToken());
			ecut[mediacur]=Double.parseDouble(st.nextToken());
			conti[mediacur]=Integer.parseInt(st.nextToken())>0?true:false;
			rho[mediacur]=Double.parseDouble(st.nextToken());
			taux=st.nextToken();
			while(st.hasMoreTokens()) taux=taux+" "+st.nextToken();
			med[mediacur]=taux;
		    }
		    ifile.close();
		}
		else Output.err.println("Error: File "+mediadef+" cannot be found");
	    }catch(Exception error){
		Output.err.println("Error: File "+mediadef+" is corrupt: "+error.toString());
		medianum=1; mediaseg=1; sphn[0]=true;
	    }
	    mediaopts+="HI  total number of defined media is "+medianum+"\n";
	}

	fnm = new int[mediaseg];
	fnx = new double[mediaseg];

	options="HI  length = "+Output.f(length)+" m  radius = "+Output.f(radius)+" m  medium = \""+med[0]+"\"\n";
	if(SURF){
	    options+="HI  propagating to surface z = "+Output.f(surf)+" m";
	    if(FACE){ if(surf<0) options+=" (up only)"; else if(surf>0) options+=" (down only)"; }
	    options+="\n";
	}
	for(i=0; i<medianum; i++){
	    options+="HI  vcut = "+Output.f(vcut[i])+"  ecut = "+Output.f(ecut[i])+" MeV";
	    if(i==0){
		if(SDEC) options+="  sdec";
		if(timef) options+="  time";
	    }
	    if(conti[i]) options+="  cont";
	    if(i==0){
		options+="  bs="+bspar;
		options+="  ph="+pncrs;
		options+="  bb="+pncbb;
		options+="  sh="+pncsh;
		if(lpmef) options+="  lpm";
		if(scatt) options+="  scat";
		options+="  romb="+romb;
	    }
	    else options+="  medium = \""+med[i]+"\"";
	    if(rho[i]!=1) options+="  rho="+Output.f(rho[i]);
	    options+="\n";
	}
	options+=mediaopts;

	Output.AMASIM=amasim;
	PhysicsModel.elow=elow;
	PhysicsModel.ebig=ebig;
	Propagate.g=romb;
	if(SEED){
	    options+="HI  random number generator seed is set at "+seed+"\n";
	    Propagate.rand.setSeed(seed);
	}
	Output.raw=raw;
	Output.err.print(options);

	p1 = new Propagate[medianum];
	p2 = new Propagate[medianum];
	p3 = new Propagate[medianum];

	String settings[] = new String[medianum];

	for(i=0; i<medianum; i++){
	    settings[i]=Output.f(vcut[i])+" "+Output.f(ecut[i])+" "+Output.f(rho[i])+" "+med[i].toLowerCase()+" "+conti[i];
	    for(j=0; j<i; j++) if(settings[j].equals(settings[i])) break;
	    if(i==j){
		double vaux=this.vaux, eaux=this.eaux;
		if(vcut[i]<0){ eaux=-vcut[i]; vcut[i]=-1; }
		if(ecut[i]<0){ vaux=-ecut[i]; ecut[i]=-1; }

		p1[i] = new Propagate(med[i], eaux, vcut[i], muta, rho[i]);
		p2[i] = new Propagate(med[i], ecut[i], vaux, muta, rho[i]);
		p3[i] = new Propagate(med[i], -1.0, -1.0, muta, rho[i]);

		p1[i].contiCorr=conti[i]; // To randomize the continuous energy losses, use only for small vcut
		p1[i].exactTime=timef;    // To compute local time of the particle exactly
		p1[i].s.lpm=lpmef;        // Enable lpm and dielectric suppression effects
		p1[i].s.b.form=bspar;     // Choose parametrization of the bremsstrahlung cross section
		p1[i].s.n.form=pncrs;     // Choose parametrization of the photon-nucleon cross section
		p1[i].s.n.bb=pncbb;       // Choose parametrization of the photon-nucleon cross section
		p1[i].s.n.shadow=pncsh;   // Choose parametrization of the photon-nucleon cross section
		p1[i].molieScat=scatt;    // To enable Moliere scattering of the muon
		p1[i].s.ci=crsci;         // Cross section multiplicative modifier
		p1[i].s.cb=crscb;         // Cross section multiplicative modifier
		p1[i].s.cp=crscp;         // Cross section multiplicative modifier
		p1[i].s.ce=crsce;         // Cross section multiplicative modifier
		p1[i].s.cd=crscd;         // Cross section multiplicative modifier
		p2[i].sdec=SDEC;          // To enable stopped muon decay
		p2[i].exactTime=timef;    // To compute local time of the particle exactly
		p2[i].s.lpm=lpmef;        // Enable lpm and dielectric suppression effects
		p2[i].s.b.form=bspar;     // Choose parametrization of the bremsstrahlung cross section
		p2[i].s.n.form=pncrs;     // Choose parametrization of the photon-nucleon cross section
		p2[i].s.n.bb=pncbb;       // Choose parametrization of the photon-nucleon cross section
		p2[i].s.n.shadow=pncsh;   // Choose parametrization of the photon-nucleon cross section
		p2[i].molieScat=scatt;    // To enable Moliere scattering of the muon
		p2[i].s.ci=crsci;         // Cross section multiplicative modifier
		p2[i].s.cb=crscb;         // Cross section multiplicative modifier
		p2[i].s.cp=crscp;         // Cross section multiplicative modifier
		p2[i].s.ce=crsce;         // Cross section multiplicative modifier
		p2[i].s.cd=crscd;         // Cross section multiplicative modifier
		p3[i].s.lpm=lpmef;        // Enable lpm and dielectric suppression effects
		p3[i].s.b.form=bspar;     // Choose parametrization of the bremsstrahlung cross section
		p3[i].s.n.form=pncrs;     // Choose parametrization of the photon-nucleon cross section
		p3[i].s.n.bb=pncbb;       // Choose parametrization of the photon-nucleon cross section
		p3[i].s.n.shadow=pncsh;   // Choose parametrization of the photon-nucleon cross section
		p3[i].s.ci=crsci;         // Cross section multiplicative modifier
		p3[i].s.cb=crscb;         // Cross section multiplicative modifier
		p3[i].s.cp=crscp;         // Cross section multiplicative modifier
		p3[i].s.ce=crsce;         // Cross section multiplicative modifier
		p3[i].s.cd=crscd;         // Cross section multiplicative modifier

		p1[i].interpolate("all", tdir+dtct);
		if(sphn[i]) p2[i].interpolate("all", tdir+dtct);
		p3[i].interpolate("all", tdir+dtct);
	    }
	    else{
		p1[i]=p1[j];
		p2[i]=p2[j];
		p3[i]=p3[j];
		if(sphn[i]) if(!p2[i].jt) p2[i].interpolate("all", tdir+dtct);
	    }
	}
	return true;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * Main parser for the f2k file streams.
     */

    private void mmcf2k(String[] args){
	Output.I3flag=false; dtct=".amanda";
	if(!setp(args)) return;

	Output.err.println(" ---  *** Enter your input in F2000 format now ***  --- ");

	try{
	    LineNumberReader file = new LineNumberReader(new InputStreamReader(Output.in));
	    StringTokenizer st;
	    String line, taux;
	    Vector buffer = new Vector(Output.HISTSIZE);
	    hist = new StringBuffer(Output.HISTSIZE);
	    userbf = new Vector(Output.HISTSIZE);
	    boolean tbegin=false;

	    int i, j, ipar;

	    if((line=file.readLine())!=null){
		Output.out.println(line);
		Output.out.println("HI  "+Output.version+"\nHI  "+param);
		Output.out.print(options);
		Output.out.println("HI  All "+muta+"'s must have energies from "+
				   Output.f(p1[0].p.low*1.e-3)+" GeV to "+Output.f(PhysicsModel.ebig*1.e-3)+" GeV");
		if(USFI) options=" for fixed z = "+Output.f(zset)+" m"; else options="";
		if(USER) Output.out.println("HI  User line mmc_en will be enabled"+options);
		else{ Output.out.println("HI  Use \"-user\" option to enable mmc_en user line"); rfix=false; }
		if(rfix) Output.out.println("HI  enforcing compliance with rdmc: exactly one mmc_en user line per event");
		if(amasim) Output.out.println("HI  some particles will be substituted with AMASIM-compatible types");
	    }

	    while((line=file.readLine())!=null){
		st = new StringTokenizer(line);
		if(!st.hasMoreTokens()) continue;
		taux=st.nextToken();
		if("TBEGIN".equals(taux)){
		    if(USER) Output.out.println("USER_DEF mmc_en NR E_INI E_CPD E_IN E_OUT CDP_X CDP_Y CDP_Z Z_IN Z_OUT");
		    Output.out.println(line);
		    tbegin=true;
		    continue;
		}
		if(!tbegin){
		    Output.out.println(line);
		    continue;
		}
		buffer.addElement(line);
		if("TR".equals(taux)){
		    igen=Integer.parseInt(st.nextToken());
		    if(igen>gens) gens=igen;
		    prnt=st.nextToken();
		    try{
			ipar=Integer.parseInt(prnt);
			if(ipar>gens) gens=ipar;
		    }catch(NumberFormatException error){
			ipar=-1;
		    }
		}
		else if("EE".equals(taux) || "END".equals(taux)){
		    if(USER){ userbf.clear(); imax=0; emax=0; }

		    for(i=0; i<buffer.size(); i++){
			line=(String)buffer.elementAt(i);
			st = new StringTokenizer(line);
			taux=st.nextToken();
			if("TR".equals(taux)){

			    tracks++;
			    igen=Integer.parseInt(st.nextToken());
			    prnt=st.nextToken();
			    type=st.nextToken();
			    if(muta.equals(type) || (muta+"-").equals(type) || (muta+"+").equals(type)){

				double x, y, z, th, phi, l, e, t;

				x=Double.parseDouble(st.nextToken());
				y=Double.parseDouble(st.nextToken());
				z=Double.parseDouble(st.nextToken());
				th=Double.parseDouble(st.nextToken());
				phi=Double.parseDouble(st.nextToken());
				if(lfix){
				    st.nextToken();
				    l=0;
				}
				else try{
				    l=Double.parseDouble(st.nextToken());
				    if(l<0) l=0;
				    else l*=1.e2;
				}catch(NumberFormatException error){
				    l=0;
				}
				e=Double.parseDouble(st.nextToken());
				t=Double.parseDouble(st.nextToken());

				prop(x, y, z, th, phi, l, e, t);

			    }
			    else{
				type="TR "+igen+" "+prnt+" "+type+st.nextToken("\n");
				Output.out.println(type);
			    }
			}
			else if("EM".equals(taux)){
			    events++;
			    Output.out.println(line);
			}
			else if("EE".equals(taux)){
			    if(USER){
				if(rfix){
				    if(userbf.size()==0) Output.out.println("US mmc_en ? ? ? ? ? ? ? ? ? ?");
				    else Output.out.println("US mmc_en "+userbf.elementAt(imax));
				}
				else for(j=0; j<userbf.size(); j++) Output.out.println("US mmc_en "+userbf.elementAt(j));
			    }
			    Output.out.println(line);
			}
			else Output.out.println(line);
		    }

		    gens=0;
		    buffer.clear();
		}

	    }

	    if(events!=0) vertices/=events;
	    Output.err.println("events read "+events+" tracks looped "+tracks+
			       " average vertices "+vertices+" missed volume "+missed);
	}catch(Exception error){
	    Output.err.println("Program finished with exception: "+error.toString());
	    throw new mmcException("input error");
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * Main propagator routine.
     */

    private double prop(double x, double y, double z, double th, double phi, double l, double e, double t){
	boolean mflag;
	int j, m, mi, mf, md, mj;

	Propagate pp=null;
	double b, d, r1=0, r2=0, r3, rr, h3, n2=0, aux;
	double result;

	double sinth, sinph, costh, cosph;

	sinph=Math.sin(phi*Math.PI/180);
	cosph=Math.cos(phi*Math.PI/180);
	sinth=Math.sin(th*Math.PI/180);
	costh=Math.cos(th*Math.PI/180);

	b=x*cosph+y*sinph;
	d=b*b+radius*radius-x*x-y*y;
	if(d>0){
	    d=Math.sqrt(d);
	    if(costh!=0){
		h1=(z-length/2)/costh;
		h2=(z+length/2)/costh;
		if(h1>h2){ aux=h1; h1=h2; h2=aux; }
	    }
	    if(sinth!=0){
		r1=(b-d)/sinth;
		r2=(b+d)/sinth;
		if(r1>r2){ aux=r1; r1=r2; r2=aux; }
		if(costh==0){
		    if(z>-length/2 && z<length/2){ h1=r1; h2=r2; }
		    else{ h1=0; h2=0; }
		}
		else{
		    if(h1>=r2 || h2<=r1){ h1=0; h2=0; }
		    else{ h1=Math.max(r1, h1); h2=Math.min(r2, h2); }
		}
	    }
	}
	else{ h1=0; h2=0; }
	if(h1<0) h1=0;
	if(h2<0) h2=0;
	if(h1==h2) missed++;
	h3=BIG;

	if(SURF){
	    if(costh!=0) aux=(z-surf)/costh;
	    else aux=BIG;
	    if(aux<0) aux=BIG;
	    if(FACE) if((z-surf)*surf<0) aux=0;
	    if(h1>aux) h1=aux;
	    if(h2>aux) h2=aux;
	    if(h3>aux) h3=aux;
	}

	if(USER){
	    if(USFI){
		if(costh!=0) n2=(z-zset)/costh;
		else n2=0;
	    }
	    else n2=x*sinth*cosph+y*sinth*sinph+z*costh;
	    if(n2<h1) flag=1;
	    else if(n2<h2) flag=2;
	    else if(n2<h3) flag=3;
	    else flag=0;
	}

	fnu=1;
	fnm[0]=0;
	fnx[0]=BIG;
	for(m=1; m<medianum; m++){
	    mflag=false;
	    if(spht[m]==1){
		b=(x*cosph+y*sinph)*sinth+(z-sphz[m]+sphr[m])*costh;
		d=b*b-(x*x+y*y+(z-sphz[m])*(z-sphz[m]+2*sphr[m]));
		if(d>0){
		    d=Math.sqrt(d);
		    r1=b-d;
		    r2=b+d;
		    mflag=true;
		}
	    }
	    if(mflag){
		md=0;
		mi=0; if(r1>0){ md++; for(mi=0; mi<fnu; mi++) if(fnx[mi]>r1) break; }
		mf=0; if(r2>0){ md++; for(mf=mi; mf<fnu; mf++) if(fnx[mf]>r2) break; }
		md+=mi-mf;
		if(md>0) for(mj=fnu-1; mj>=mf; mj--){
		    fnx[mj+md]=fnx[mj];
		    fnm[mj+md]=fnm[mj];
		}
		else if(md<0) for(mj=mf; mj<fnu; mj++){
		    fnx[mj+md]=fnx[mj];
		    fnm[mj+md]=fnm[mj];
		}
		fnu+=md;
		if(r1>0){
		    fnx[mi]=r1;
		    mi++;
		}
		if(r2>0){
		    fnx[mi]=r2;
		    fnm[mi]=m;
		}
	    }
	}
	
	if(Output.I3flag) I3hist.clear(); else hist.setLength(0);
	if(USER){ e1=0; e2=0; ec=0; }
	pp=null; result=e; rr=0;

	for(j=0; j<fnu; j++){
	    m=fnm[j]; aux=fnx[j];
	    r1=Math.min(h1, aux);
	    r2=Math.min(h2, aux);
	    r3=Math.min(h3, aux);
	    if(pp==null) result=propagateTo(r1, e, n2, 1, p1[m], 0, 0,
					    t*1.e-9, x*1.e2, y*1.e2, z*1.e2, 180-th, phi<180?phi+180:phi-180);
	    else result=propagateTo(r1-rr, result, n2-rr, 1, p1[m], 0, 0,
				    pp.p.t, pp.p.x, pp.p.y, pp.p.z, pp.p.theta, pp.p.phi);
	    pp=p1[m]; l+=pp.p.r; if(USER) if(r1>=rr && r1==h1) e1=result;
	    if(result>0){
		if(r1>rr) rr=r1;
		result=propagateTo(r2-rr, result, n2-rr, 2, p2[m], igen, gens,
				   pp.p.t, pp.p.x, pp.p.y, pp.p.z, pp.p.theta, pp.p.phi);
		pp=p2[m]; l+=pp.p.r; if(USER) if(r2>=rr && r2==h2) e2=result;
		if(Output.I3flag) I3hist.addAll(pp.o.I3hist); else hist.append(pp.o.history);
		vertices+=pp.o.gens-gens; gens=pp.o.gens;
		if(result>0){
		    if(r2>rr) rr=r2;
		    result=propagateTo(r3-rr, result, n2-rr, 3, p3[m], 0, 0,
				       pp.p.t, pp.p.x, pp.p.y, pp.p.z, pp.p.theta, pp.p.phi);
		    pp=p3[m]; l+=pp.p.r;
		    if(result>0) if(r3>rr) rr=r3;
		}
	    }
	    if(result<=0) break;
	}

	if(USER){
	    if(h1==0) e1=0;
	    if(h2==0) e2=0;
	    if(e1<0) e1-=rr;
	    if(e2<0) e2-=rr;
	    z1=z-costh*h1;
	    z2=z-costh*h2;
	    nx=x-sinth*cosph*n2;
	    ny=y-sinth*sinph*n2;
	    nz=z-costh*n2;
	    if(!Output.I3flag){
		userbf.addElement(igen+" "+Output.f(e)+" "+(ec>0?Output.f(ec):"?")+" "+
				  (e1!=0?Output.f(e1):"?")+" "+(e2!=0?Output.f(e2):"?")+" "+
				  Output.f(nx)+" "+Output.f(ny)+" "+Output.f(nz)+" "+
				  (h1>0?Output.f(z1):"?")+" "+(h2>0?Output.f(z2):"?"));
		if(rfix) if(e>emax){ imax=userbf.size()-1; emax=e; }
	    }
	}

	if(SURF){
	    x=pp.p.x*1.e-2;
	    y=pp.p.y*1.e-2;
	    z=pp.p.z*1.e-2;
	    th=180-pp.p.theta;
	    phi=pp.p.phi; phi=phi<180?phi+180:phi-180;
	    if(result>0){ e=result; l=-BIG; }
	    else{ e=0; l=0; }
	    t=pp.p.t*1.e9;
	}
	if(!Output.I3flag){
	    Output.out.println("TR "+igen+" "+prnt+" "+type+" "+Output.f(x)+" "+Output.f(y)+" "+Output.f(z)+" "+
			       Output.f(th)+" "+Output.f(phi)+" "+(l>=0?Output.f(l*1.e-2):"?")+" "+
			       Output.f(e)+" "+Output.f(t));
	    Output.out.print(hist);
	}

	return l;
    }

    //----------------------------------------------------------------------------------------------------//

    private int flag=0;

    /**
     * user-block variable: z-coordinate of entry point into the detector cylinder (Z_IN)
     */

    public double z1=0;

    /**
     * user-block variable: z-coordinate of exit point from the detector cylinder (Z_OUT)
     */

    public double z2=0;

    /**
     * user-block variable: distance to the point of entry into the detector cylinder (R_IN)
     */

    public double h1=0;

    /**
     * user-block variable: distance to the point of exit from the detector cylinder (R_OUT)
     */

    public double h2=0;

    /**
     * user-block variable: x-coordinate of closest approach point (CPD_X) or -user=z point
     */

    public double nx=0;

    /**
     * user-block variable: y-coordinate of closest approach point (CPD_Y) or -user=z point
     */

    public double ny=0;

    /**
     * user-block variable: z-coordinate of closest approach point (CPD_Z) or -user=z point
     */

    public double nz=0;

    /**
     * user-block variable: Energy at the entry point into the detector cylinder if positive (E_IN)
     */

    public double e1=0;

    /**
     * user-block variable: Energy at the exit point from the detector cylinder if positive (E_OUT)
     */

    public double e2=0;

    /**
     * user-block variable: Energy at the CPD point if positive (E_CPD)
     */

    public double ec=0;

    //----------------------------------------------------------------------------------------------------//

    /**
     * checks the flag, corrects the units, initializes the particle and calls the propagator
     */

    private double propagateTo(double h, double e, double n, int i, Propagate p, int igen, int gens,
				      double time, double x, double y, double z, double theta, double phi){
	double result;
	switch(i){
	case 1: p.o.initDefault(type, time, x, y, z, theta, phi); break;
	case 2: p.o.initF2000(igen, gens, type, time, x, y, z, theta, phi); break;
	case 3: p.o.initDefault(type, time, x, y, z, theta, phi); break;
	}
	if(flag==i && n>=0 && n<h){
	    result=p.propagateTo(h*1.e2, e*1.e3, n*1.e2);
	    ec=p.getPropEc();
	    ec=ec>0?ec*1.e-3:ec*1.e-2;
	}
	else result=p.propagateTo(h*1.e2, e*1.e3);
	return result>0?result*1.e-3:result*1.e-2;
    }

}
